﻿/* 
值动画管理器
作者：滴风
修改日期：2023.12
用法写在该文件末尾
-----------------------------------------
修改：支持延时动画和分段动画
修改：支持贝塞尔曲线动画
修改：添加清理功能解决场景退出闪烁问题
*/

class ValueAnimateBezierVector3 : ValueAnimate {
    Vector3 beginValue;
    Vector3 controlPoint;
    Vector3 endValue;
    func_apply_vector3_t@ callback;
    private Vector3 currentValue;

    bool Update(float timeStep) {
        // 必须调用父类的Update获取基础时间处理
        if (!ValueAnimate::Update(timeStep)) 
            return false; // 父类已处理延时/结束状态
            
        float t = elapsedTime / duration;
        currentValue = _BezierCalc(beginValue, controlPoint, endValue, t);
        return true; // 返回true表示需要更新
    }

    private Vector3 _BezierCalc(const Vector3& p0, const Vector3& p1, const Vector3& p2, float t) {
        float u = 1.0 - t;
        return Vector3(
            (u*u)*p0.x + (2*u*t)*p1.x + (t*t)*p2.x,
            (u*u)*p0.y + (2*u*t)*p1.y + (t*t)*p2.y,
            (u*u)*p0.z + (2*u*t)*p1.z + (t*t)*p2.z
        );
    }

    void Apply(){
        if (callback !is null){
            callback(currentValue);
        }
    }
}

abstract class ValueAnimate{
    float duration;
    float delay = 0.0;         // 动画开始前的延时时间
    protected float elapsedTime;
    protected float delayTimer = 0.0;  // 延时计时器

    // 返回true表示值有变化
    bool Update(float timeStep) {
        // 处理延时阶段
        if (delayTimer < delay) {
            delayTimer += timeStep;
            return false;
        }
        
        float t = elapsedTime / duration;
        if (t > 1){
            return false;
        }
        elapsedTime += timeStep;
        return true;
    }
    
    bool IsEnd() {
        return (delayTimer >= delay) && (elapsedTime > duration);
    }
    
    void Apply() {}
}

//-----------------------------------------
// 以下是具体动画类型实现
//-----------------------------------------

funcdef void func_apply_float_t(float value);
funcdef void func_apply_vector2_t(Vector2 value);
funcdef void func_apply_vector3_t(Vector3 value);

class ValueAnimateFloat : ValueAnimate{
    float beginValue;
    float endValue;
    const func_apply_float_t@ callback;
    private float currentValue;

    bool Update(float timeStep) {
        if (!ValueAnimate::Update(timeStep)) 
            return false;
            
        float t = elapsedTime / duration;
        currentValue = Lerp(beginValue, endValue, t);
        return true;
    }

    void Apply(){
        if (callback !is null){
            callback(currentValue);
        }
    }
}

//-----------------------------------------

Vector2 _Lerp(const Vector2& lhs, const Vector2& rhs, float t) { 
    return lhs * (1.0 - t) + rhs * t; 
}

class ValueAnimateVector2 : ValueAnimate{
    Vector2 beginValue;
    Vector2 endValue;
    const func_apply_vector2_t@ callback;
    private Vector2 currentValue;

    bool Update(float timeStep){
        if (!ValueAnimate::Update(timeStep)) 
            return false;
            
        float t = elapsedTime / duration;
        currentValue = _Lerp(beginValue, endValue, t);
        return true;
    }

    void Apply(){
        if (callback !is null){
            callback(currentValue);
        }
    }
}

//-----------------------------------------

Vector3 _Lerp(const Vector3& lhs, const Vector3& rhs, float t) { 
    return lhs * (1.0 - t) + rhs * t; 
}

class ValueAnimateVector3 : ValueAnimate {
    Vector3 beginValue;
    Vector3 endValue;
    func_apply_vector3_t@ callback;
    private Vector3 currentValue;

    bool Update(float timeStep){
        if (!ValueAnimate::Update(timeStep)) 
            return false;
            
        float t = elapsedTime / duration;
        currentValue = _Lerp(beginValue, endValue, t);
        return true;
    }

    void Apply(){
        if (callback !is null){
            callback(currentValue);
        }
    }
}

//-----------------------------------------
// 动画管理器（添加清理功能）
//-----------------------------------------

class ValueAnimateManager{
    private ValueAnimate@[] valueAnimateList;
    private float lastGameTime;

    // 新增：清理所有动画
    void ClearAll() {
        valueAnimateList.Clear();
        lastGameTime = 0;  // 重置时间戳
    }

    void Add(ValueAnimate@ valueAnimate){
        valueAnimateList.Push(valueAnimate);
    }

    void Update(){
        if (lastGameTime == 0){
            lastGameTime = GetGameTimeFloat();
        }
        float currentGameTime = GetGameTimeFloat();
        float timeStep = currentGameTime - lastGameTime;
        
        uint invalidCount = 0;
        for (uint i = 0; i < valueAnimateList.GetSize(); i++){
            auto@ valueAnimate = valueAnimateList[i];
            if (valueAnimate.Update(timeStep)){
                valueAnimate.Apply();
            }
            if (valueAnimate.IsEnd()){
                invalidCount += 1;
            }
        }

        // 定期清理已完成的动画
        if (invalidCount > 10 || invalidCount == valueAnimateList.GetSize()) {
            ValueAnimate@[] validList;
            for (uint i = 0; i < valueAnimateList.GetSize(); i++){
                auto@ valueAnimate = valueAnimateList[i];
                if (!valueAnimate.IsEnd()){
                    validList.Push(valueAnimate);
                }
            }
            valueAnimateList.Swap(validList);
        }
        
        lastGameTime = currentGameTime;
    }
}

ValueAnimateManager g_valueAnimateManager;